@@ -1,7 +1,7 @@
-/*
+ /*
  *  arch-tag: Header for RhythmDB - Rhythmbox backend queryable database
  *
- *  Copyright (C) 2003 Colin Walters <walters@rhythmbox.org>
+ *  Copyright (C) 2003,2004 Colin Walters <walters@rhythmbox.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -26,6 +26,11 @@
 #include <glib-object.h>
 #include <gtk/gtktreemodel.h>
 #include <stdarg.h>
+#include <libgnomevfs/gnome-vfs-utils.h>
+#include <libgnomevfs/gnome-vfs-file-info.h>
+#include <libxml/tree.h>
+
+#include "config.h"
 
 G_BEGIN_DECLS
 
@@ -36,50 +41,70 @@ G_BEGIN_DECLS
 #define RHYTHMDB_IS_CLASS(k)  (G_TYPE_CHECK_CLASS_TYPE ((k), RHYTHMDB_TYPE))
 #define RHYTHMDB_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), RHYTHMDB_TYPE, RhythmDBClass))
 
-typedef enum
-{
-	RHYTHMDB_ENTRY_TYPE_SONG,
-	RHYTHMDB_ENTRY_TYPE_IRADIO_STATION
-} RhythmDBEntryType;
+
+typedef gint32 RhythmDBEntryType;
+
+#define RHYTHMDB_ENTRY_TYPE_SONG (rhythmdb_entry_song_get_type ())
+#define RHYTHMDB_ENTRY_TYPE_IRADIO_STATION (rhythmdb_entry_iradio_get_type ())
+
 
 typedef enum
 {
 	RHYTHMDB_QUERY_END,
 	RHYTHMDB_QUERY_DISJUNCTION,
+	RHYTHMDB_QUERY_SUBQUERY,
 	RHYTHMDB_QUERY_PROP_EQUALS,
 	RHYTHMDB_QUERY_PROP_LIKE,
+	RHYTHMDB_QUERY_PROP_NOT_LIKE,
 	RHYTHMDB_QUERY_PROP_GREATER,
 	RHYTHMDB_QUERY_PROP_LESS,
 } RhythmDBQueryType;
 
-#define RHYTHMDB_NUM_PROPERTIES 19
-#define RHYTHMDB_NUM_SAVED_PROPERTIES 14
-
+/* If you modify this enum, don't forget to modify rhythmdb_prop_get_type */
 typedef enum
 {
-	RHYTHMDB_PROP_TYPE,
-	RHYTHMDB_PROP_NAME,
+	RHYTHMDB_PROP_TYPE = 0,
+	RHYTHMDB_PROP_TITLE,
 	RHYTHMDB_PROP_GENRE,
 	RHYTHMDB_PROP_ARTIST,
 	RHYTHMDB_PROP_ALBUM,
 	RHYTHMDB_PROP_TRACK_NUMBER,
+	RHYTHMDB_PROP_DISC_NUMBER,
 	RHYTHMDB_PROP_DURATION,
 	RHYTHMDB_PROP_FILE_SIZE,
 	RHYTHMDB_PROP_LOCATION,
 	RHYTHMDB_PROP_MTIME,
 	RHYTHMDB_PROP_RATING,
+	RHYTHMDB_PROP_AUTO_RATE,
 	RHYTHMDB_PROP_PLAY_COUNT,
 	RHYTHMDB_PROP_LAST_PLAYED,
-	RHYTHMDB_PROP_QUALITY,
+	RHYTHMDB_PROP_BITRATE,
+	RHYTHMDB_PROP_TRACK_GAIN,
+	RHYTHMDB_PROP_TRACK_PEAK,
+	RHYTHMDB_PROP_ALBUM_GAIN,
+	RHYTHMDB_PROP_ALBUM_PEAK,
+	RHYTHMDB_PROP_MIMETYPE,
+	/* This should always be the last element of this enum */
+	RHYTHMDB_NUM_SAVED_PROPERTIES
 } RhythmDBPropType;
 
+
+/* If you modify this enum, don't forget to modify 
+ * rhythmdb_unsaved_prop_get_type 
+ */
 typedef enum
 {
-	RHYTHMDB_PROP_NAME_SORT_KEY = 14,
+	RHYTHMDB_PROP_TITLE_SORT_KEY = RHYTHMDB_NUM_SAVED_PROPERTIES,
 	RHYTHMDB_PROP_GENRE_SORT_KEY,
 	RHYTHMDB_PROP_ARTIST_SORT_KEY,
 	RHYTHMDB_PROP_ALBUM_SORT_KEY,
+	RHYTHMDB_PROP_TITLE_FOLDED,
+	RHYTHMDB_PROP_GENRE_FOLDED,
+	RHYTHMDB_PROP_ARTIST_FOLDED,
+	RHYTHMDB_PROP_ALBUM_FOLDED,
 	RHYTHMDB_PROP_LAST_PLAYED_STR,
+	/* This should always be the last element of this enum */
+	RHYTHMDB_NUM_PROPERTIES
 } RhythmDBUnsavedPropType;
 
 GType rhythmdb_query_get_type (void);
@@ -94,10 +119,20 @@ typedef struct {
 	guint type;
 	guint propid;
 	GValue *val;
+	GPtrArray *subquery;
 } RhythmDBQueryData;
 
 typedef void RhythmDBEntry;
 
+typedef enum
+{
+	RHYTHMDB_ERROR_ACCESS_FAILED,
+} RhythmDBError;
+
+#define RHYTHMDB_ERROR (rhythmdb_error_quark ())
+
+GQuark rhythmdb_error_quark (void);
+
 typedef struct RhythmDBPrivate RhythmDBPrivate;
 
 typedef struct
@@ -113,37 +148,38 @@ typedef struct
 
 	/* signals */
 	void	(*entry_added)		(RhythmDBEntry *entry);
-
-	void	(*genre_added)		(RhythmDBEntry *entry, const char *genre);
-	void	(*artist_added)		(RhythmDBEntry *entry, const char *artist);
-	void	(*album_added)		(RhythmDBEntry *entry, const char *album);
-	void	(*genre_deleted)	(RhythmDBEntry *entry, const char *genre);
-	void	(*artist_deleted)	(RhythmDBEntry *entry, const char *artist);
-	void	(*album_deleted)	(RhythmDBEntry *entry, const char *album);
+	void	(*entry_restored)	(RhythmDBEntry *entry);
+	void	(*entry_changed)	(RhythmDBEntry *entry, RhythmDBPropType prop,
+					 const GValue *old, const GValue *new);
+	void	(*entry_deleted)	(RhythmDBEntry *entry);
+	void	(*load_complete)	(RhythmDBEntry *entry);
+	void	(*error)		(const char *uri, const char *msg);
 
 	/* virtual methods */
 
-	void		(*impl_load)		(RhythmDB *db, GMutex *mutex, gboolean *dead);
+	void		(*impl_load)		(RhythmDB *db, gboolean *dead);
 	void		(*impl_save)		(RhythmDB *db);
 	
 	RhythmDBEntry *	(*impl_entry_new)	(RhythmDB *db, RhythmDBEntryType type,
 						 const char *uri);
 
 	void		(*impl_entry_set)	(RhythmDB *db, RhythmDBEntry *entry,
-						 guint propid, GValue *value);
+						 guint propid, const GValue *value);
 
 	void		(*impl_entry_get)	(RhythmDB *db, RhythmDBEntry *entry,
 						 guint propid, GValue *value);
 
 	void		(*impl_entry_delete)	(RhythmDB *db, RhythmDBEntry *entry);
+
+	void            (*impl_entry_delete_by_type) (RhythmDB *db, RhythmDBEntryType type);
+
 	RhythmDBEntry *	(*impl_lookup_by_location)(RhythmDB *db, const char *uri);
 
-	void		(*impl_do_full_query)	(RhythmDB *db, GPtrArray *query,
-						 GtkTreeModel **main_model,
-						 GtkTreeModel **genre_model,
-						 GtkTreeModel **artist_model,
-						 GtkTreeModel **album_model);
+	gboolean 	(*impl_evaluate_query)	(RhythmDB *db, GPtrArray *query, RhythmDBEntry *entry);
 
+	void		(*impl_do_full_query)	(RhythmDB *db, GPtrArray *query,
+						 GtkTreeModel *main_model,
+						 gboolean *cancel);
 } RhythmDBClass;
 
 GType		rhythmdb_get_type	(void);
@@ -157,8 +193,6 @@ void		rhythmdb_shutdown	(RhythmDB *db);
  */
 void		rhythmdb_load		(RhythmDB *db);
 
-void		rhythmdb_load_join	(RhythmDB *db);
-
 void		rhythmdb_save		(RhythmDB *db);
 
 void		rhythmdb_read_lock	(RhythmDB *db);
@@ -166,15 +200,35 @@ void		rhythmdb_write_lock	(RhythmDB *db);
 void		rhythmdb_read_unlock	(RhythmDB *db);
 void		rhythmdb_write_unlock	(RhythmDB *db);
 
+gboolean	rhythmdb_entry_is_editable (RhythmDB *db, RhythmDBEntry *entry);
+
 RhythmDBEntry *	rhythmdb_entry_new	(RhythmDB *db, RhythmDBEntryType type, const char *uri);
 
+void		rhythmdb_add_uri_async	(RhythmDB *db, const char *uri);
+RhythmDBEntry *	rhythmdb_add_song	(RhythmDB *db, const char *uri, GError **error);
+
 void		rhythmdb_entry_set	(RhythmDB *db, RhythmDBEntry *entry,
 					 guint propid, GValue *value);
 
+void		rhythmdb_entry_queue_set(RhythmDB *db, RhythmDBEntry *entry,
+					 guint propid, GValue *value);
+
 void		rhythmdb_entry_get	(RhythmDB *db, RhythmDBEntry *entry,
 					 guint propid, GValue *value);
 
+#ifndef WITH_RHYTHMDB_TREE
+#define rhythmdb_entry_ref(DB, ENTRY) 
+#define rhythmdb_entry_ref_unlocked(DB, ENTRY) 
+#define rhythmdb_entry_unref(DB, ENTRY) 
+#define rhythmdb_entry_unref_unlocked(DB, ENTRY) 
+#else
+#include "rhythmdb-tree.h"
+#endif
+
 void		rhythmdb_entry_delete	(RhythmDB *db, RhythmDBEntry *entry);
+void            rhythmdb_entry_delete_by_type (RhythmDB *db, 
+					       RhythmDBEntryType type);
+
 
 const char *	rhythmdb_entry_get_string	(RhythmDB *db,
 						 RhythmDBEntry *entry,
@@ -185,6 +239,9 @@ gboolean	rhythmdb_entry_get_boolean	(RhythmDB *db,
 long		rhythmdb_entry_get_long		(RhythmDB *db,
 						 RhythmDBEntry *entry,
 						 guint property_id);
+guint64		rhythmdb_entry_get_uint64	(RhythmDB *db,
+						 RhythmDBEntry *entry,
+						 guint property_id);
 int		rhythmdb_entry_get_int		(RhythmDB *db,
 						 RhythmDBEntry *entry,
 						 guint property_id);
@@ -200,44 +257,64 @@ gpointer	rhythmdb_entry_get_pointer	(RhythmDB *db,
 
 RhythmDBEntry *	rhythmdb_entry_lookup_by_location (RhythmDB *db, const char *uri);
 
+gboolean	rhythmdb_evaluate_query		(RhythmDB *db, GPtrArray *query,
+						 RhythmDBEntry *entry);
+
 /**
  * Returns a freshly allocated GtkTreeModel which represents the query.
  * The extended arguments alternate between RhythmDBQueryType args
  * and their values.  Here's an example:
  *
- * rhythmdb_do_entry_query (db, RHYTHMDB_QUERY_PROP_EQUALS, "genre", "Classical",
+ * rhythmdb_do_full_query (db, RHYTHMDB_QUERY_PROP_EQUALS, "genre", "Classical",
  *                          RHYTHMDB_QUERY_PROP_GREATER, "rating", 5,
  *                          RHYTHMDB_QUERY_END);
  *
- * Actually, you wouldn't literally specify "Classical" as above.  More on that
- * issue below.
- *
  */
-/* GtkTreeModel *	rhythmdb_do_entry_query			(RhythmDB *db, ...); */
-
-/* This is a specialized query to return a flat list of metadata,
- * e.g. genre/artist/album.  The varargs are the same as for the
- * rhythmdb_do_entry_query call.  You should think of it, conceptually,
- * as constructing a list of all entries which match the query, and then
- * returning the values of the attribute as a set.
- */
-/* GtkTreeModel *	rhythmdb_do_property_query		(RhythmDB *db, guint property_id, ...); */
-
 void		rhythmdb_do_full_query			(RhythmDB *db,
-							 GtkTreeModel **main_model,
-							 GtkTreeModel **genre_model,
-							 GtkTreeModel **artist_model,
-							 GtkTreeModel **album_model, ...);
+							 GtkTreeModel *main_model,
+							 ...);
+void		rhythmdb_do_full_query_parsed		(RhythmDB *db,
+							 GtkTreeModel *main_model,
+							 GPtrArray *query);
+
+void		rhythmdb_do_full_query_async		(RhythmDB *db, GtkTreeModel *main_model, ...);
+
+void		rhythmdb_do_full_query_async_parsed	(RhythmDB *db, GtkTreeModel *main_model,
+							 GPtrArray *query);
+
+void		rhythmdb_query_cancel			(RhythmDB *db, GtkTreeModel *query_model);
 
 GType		rhythmdb_get_property_type		(RhythmDB *db, guint property_id);
 
-void		rhythmdb_emit_genre_deleted		(RhythmDB *db, const char *genre);
-void		rhythmdb_emit_artist_deleted		(RhythmDB *db, const char *artist);
-void		rhythmdb_emit_album_deleted		(RhythmDB *db, const char *album);
+void		rhythmdb_entry_sync_mirrored		(RhythmDB *db, RhythmDBEntry *entry,
+							 guint propid, GValue *value);
+
+GPtrArray *	rhythmdb_query_parse			(RhythmDB *db, ...);
+void		rhythmdb_query_append			(RhythmDB *db, GPtrArray *query, ...);
+void		rhythmdb_query_free			(GPtrArray *query);
+GPtrArray *	rhythmdb_query_copy			(GPtrArray *array);
+
+void		rhythmdb_query_serialize		(RhythmDB *db, GPtrArray *query,
+							 xmlNodePtr node);
+
+GPtrArray *	rhythmdb_query_deserialize		(RhythmDB *db, xmlNodePtr node);
+
+inline const char *	rhythmdb_nice_elt_name_from_propid	(RhythmDB *db, gint propid);
+inline int		rhythmdb_propid_from_nice_elt_name	(RhythmDB *db, const char *name);
+
+void		rhythmdb_emit_entry_added		(RhythmDB *db, RhythmDBEntry *entry);
+void		rhythmdb_emit_entry_restored		(RhythmDB *db, RhythmDBEntry *entry);
+void		rhythmdb_emit_entry_deleted		(RhythmDB *db, RhythmDBEntry *entry);
+
+char *		rhythmdb_get_status			(RhythmDB *db);
+char *		rhythmdb_compute_status_normal		(gint n_songs, glong duration,
+							 GnomeVFSFileSize size);
+RhythmDBEntryType rhythmdb_entry_register_type          (void);
+
+RhythmDBEntryType rhythmdb_entry_song_get_type          (void);
+RhythmDBEntryType rhythmdb_entry_iradio_get_type        (void);
+
 
-void		rhythmdb_emit_genre_added		(RhythmDB *db, const char *genre);
-void		rhythmdb_emit_artist_added		(RhythmDB *db, const char *artist);
-void		rhythmdb_emit_album_added		(RhythmDB *db, const char *album);
 
 G_END_DECLS
 
