@@ -92,6 +92,7 @@ static void rb_iradio_source_songs_view_sort_order_changed_cb (RBEntryView *view
 /* source methods */
 static const char *impl_get_status (RBSource *source);
 static const char *impl_get_browser_key (RBSource *source);
+static const char *impl_get_search_key (RBSource *source);
 static GdkPixbuf *impl_get_pixbuf (RBSource *source);
 static RBEntryView *impl_get_entry_view (RBSource *source);
 static void impl_search (RBSource *source, const char *text);
@@ -115,6 +116,7 @@ void rb_iradio_source_show_columns_changed_cb (GtkToggleButton *button,
 #define CONF_STATE_PANED_POSITION CONF_PREFIX "/state/iradio/paned_position"
 #define CONF_STATE_IRADIO_SORTING CONF_PREFIX "/state/iradio/sorting"
 #define CONF_STATE_SHOW_BROWSER   CONF_PREFIX "/state/iradio/show_browser"
+#define CONF_STATE_SEARCH_TEXT   CONF_PREFIX "/state/iradio/search_text"
 
 struct RBIRadioSourcePrivate
 {
@@ -136,6 +138,8 @@ struct RBIRadioSourcePrivate
 	gboolean shuffle;
 	gboolean repeat;
 
+	gboolean initialized;
+
 	char *status;
 
 	GtkWidget *paned;
@@ -213,6 +217,7 @@ rb_iradio_source_class_init (RBIRadioSourceClass *klass)
 
 	source_class->impl_get_status  = impl_get_status;
 	source_class->impl_get_browser_key  = impl_get_browser_key;
+	source_class->impl_get_search_key  = impl_get_search_key;
 	source_class->impl_get_pixbuf  = impl_get_pixbuf;
 	source_class->impl_can_search = (RBSourceFeatureFunc) rb_true_function;
 	source_class->impl_search = impl_search;
@@ -288,7 +293,7 @@ rb_iradio_source_constructor (GType type, guint n_construct_properties,
 	RBIRadioSource *source;
 	RBIRadioSourceClass *klass;
 	GObjectClass *parent_class;  
-	klass = RB_IRADIO_SOURCE_CLASS (g_type_class_peek (type));
+	klass = RB_IRADIO_SOURCE_CLASS (g_type_class_peek (RB_TYPE_IRADIO_SOURCE));
 
 	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
 	source = RB_IRADIO_SOURCE (parent_class->constructor (type, n_construct_properties,
@@ -343,9 +348,10 @@ rb_iradio_source_constructor (GType type, guint n_construct_properties,
 	eel_gconf_notification_add (CONF_FIRST_TIME,
 				    (GConfClientNotifyFunc) rb_iradio_source_first_time_changed,
 				    source);
-	rb_iradio_source_do_query (source, RB_IRADIO_QUERY_TYPE_ALL);
-			
 	gtk_widget_show_all (GTK_WIDGET (source));
+
+	rb_iradio_source_do_query (source, RB_IRADIO_QUERY_TYPE_ALL);
+
 	return G_OBJECT (source);
 }
 
@@ -388,15 +394,21 @@ rb_iradio_source_get_property (GObject *object,
 }
 
 RBSource *
-rb_iradio_source_new (RhythmDB *db)
+rb_iradio_source_new (RBShell *shell, RhythmDB *db, 
+		      BonoboUIComponent *component)
 {
 	RBSource *source;
 
 	source = RB_SOURCE (g_object_new (RB_TYPE_IRADIO_SOURCE,
 					  "name", _("Radio"),
+					  "internal-name", "<radio>",
 					  "db", db,
+					  "component", component,
 					  NULL));
 
+	rb_shell_register_entry_type_for_source (shell, source, 
+						 RHYTHMDB_ENTRY_TYPE_IRADIO_STATION);
+
 	return source;
 }
 
@@ -428,7 +440,7 @@ rb_iradio_source_add_station (RBIRadioSource *source,
 	g_value_unset (&val);
 	
 	g_value_init (&val, G_TYPE_DOUBLE);
-	g_value_set_double (&val, 3.0);
+	g_value_set_double (&val, 2.5);
 	rhythmdb_entry_set (source->priv->db, entry, RHYTHMDB_PROP_RATING, &val);
 	g_value_unset (&val);
 	
@@ -447,12 +459,17 @@ static void
 impl_search (RBSource *asource, const char *search_text)
 {
 	RBIRadioSource *source = RB_IRADIO_SOURCE (asource);
-	if (search_text == NULL && source->priv->search_text == NULL)
-		return;
-	if (search_text != NULL &&
-	    source->priv->search_text != NULL
-	    && !strcmp (search_text, source->priv->search_text))
-		return;
+
+	if (source->priv->initialized) {
+		if (search_text == NULL && source->priv->search_text == NULL)
+			return;
+		if (search_text != NULL &&
+		    source->priv->search_text != NULL
+		    && !strcmp (search_text, source->priv->search_text))
+			return;
+	}
+
+	source->priv->initialized = TRUE;
 	if (search_text != NULL && search_text[0] == '\0')
 		search_text = NULL;
 
@@ -550,6 +567,12 @@ impl_get_browser_key (RBSource *asource)
 	return CONF_STATE_SHOW_BROWSER;
 }
 
+static const char *
+impl_get_search_key (RBSource *asource)
+{
+	return CONF_STATE_SEARCH_TEXT;
+}
+
 static void
 impl_delete (RBSource *asource)
 {
